/* ************************************************************************ 
 *         The Amulet User Interface Development Environment              *
 * ************************************************************************
 * This code was written by Paolo Cacciatori <paolo@nova.it>              *
 * as a contribution to the Amulet project, and modified by the Amulet    *
 * Group.  Amulet is a project of Carnegie Mellon University.             *
 * ************************************************************************/

// This file contains common file dialog box widgets
// Paolo Cacciatori  11-may-97
// EAB August 97
// - replaced GetINode (sequence of stat calls on "..") with getwd()
// - got rid of dead code
// - replaced definition of class DArray (redefining DynArray)
//   with class derived from DynArray, adding Sort
// - replaced in-line code with calls to common procedures
//   -- append_name_to_path
//   -- backup_pathname
// - use common constants MAXPATHLEN and MAXNAMLEN
// - use prototype objects for textLine in ReadDir
// - stat requires full pathname
// - wrote Am_Parse_Filename and Am_Directory_Contents

// move dynarray.h to include/amulet and add Sort
#define DYNARRAY__H <AMULET_ROOT##/src/object/dynarray.h>

#include <amulet.h>
#include WIDGETS_ADVANCED__H
#include STR_STREAM__H

#ifdef GCC
#include <string.h>
#include <stdlib.h>
#else
extern "C" {
#include <string.h>
#include <stdlib.h>
}
#endif
#include DYNARRAY__H

class DArray: public DynArray {
public:
  DArray (unsigned elem_size) : DynArray(elem_size) {}

  void Sort(int (*compar)(const void *, const void *)) {
    qsort(data, length, elem_size, compar);
  }
};

#if defined(_WINDOWS) 

#include <windows.h>
#include <commdlg.h>
#include <stdlib.h>
#include <stdio.h>

static OPENFILENAME ofn;

const char path_separator = '\\';
const char path_separator_string = "\\";

void PopFileInitialize (HWND hwnd)
{
     static char szFilter[] = "Text Files (*.TXT)\0*.txt\0"  \
                              "ASCII Files (*.ASC)\0*.asc\0" \
                              "All Files (*.*)\0*.*\0\0" ;

     ofn.lStructSize       = sizeof (OPENFILENAME);
     ofn.hwndOwner         = hwnd ;
     ofn.hInstance         = NULL ;
     ofn.lpstrFilter       = szFilter ;
     ofn.lpstrCustomFilter = NULL ;
     ofn.nMaxCustFilter    = 0 ;
     ofn.nFilterIndex      = 0 ;
     ofn.lpstrFile         = NULL ;          // Set in Open and Close functions
     ofn.nMaxFile          = _MAX_PATH ;
     ofn.lpstrFileTitle    = NULL ;          // Set in Open and Close functions
     ofn.nMaxFileTitle     = _MAX_FNAME + _MAX_EXT ;
     ofn.lpstrInitialDir   = NULL ;
     ofn.lpstrTitle        = NULL ;
     ofn.Flags             = 0 ;             // Set in Open and Close functions
     ofn.nFileOffset       = 0 ;
     ofn.nFileExtension    = 0 ;
     ofn.lpstrDefExt       = "txt" ;
     ofn.lCustData         = 0L ;
     ofn.lpfnHook          = NULL ;
     ofn.lpTemplateName    = NULL ;
}

int PopFileOpenDlg (HWND hwnd, PSTR pstrFileName, PSTR pstrTitleName)
{
     ofn.hwndOwner         = hwnd ;
     ofn.lpstrFile         = pstrFileName ;
     ofn.lpstrFileTitle    = pstrTitleName ;
     ofn.Flags             = OFN_HIDEREADONLY | OFN_CREATEPROMPT ;

     return GetOpenFileName (&ofn) ;
}

int PopFileSaveDlg (HWND hwnd, PSTR pstrFileName, PSTR pstrTitleName)
{
     ofn.hwndOwner         = hwnd ;
     ofn.lpstrFile         = pstrFileName ;
     ofn.lpstrFileTitle    = pstrTitleName ;
     ofn.Flags             = OFN_OVERWRITEPROMPT ;

     return GetSaveFileName (&ofn) ;
}


#elif defined(_MACINTOSH)
const char path_separator = ':';
const char path_separator_string = ":";

#else
// UNIX
#include <dirent.h>	// defines MAXNAMLEN
#include <unistd.h>
#include <sys/param.h>	// defines MAXPATHLEN
extern "C" {
#include <sys/stat.h>
}
#include <stdio.h>
#include "file_dialog_widgets.h"

Am_Object Am_GetFile_Dialog;
Am_Object Am_PutFile_Dialog;

Am_Object Dir_Line_Proto;
Am_Object Type0_File_Line_Proto;

Am_Slot_Key fileDialogType = Am_Register_Slot_Name ("fileDialog_type");
Am_Slot_Key fileDialog_txtFilter = Am_Register_Slot_Name ("fileDialog_txtFilter");
Am_Slot_Key fileDialog_txtSelection = Am_Register_Slot_Name ("fileDialog_txtSelection");
Am_Slot_Key fileDialog_ListDirs = Am_Register_Slot_Name ("fileDialog_ListDirs");
Am_Slot_Key fileDialog_ListFiles = Am_Register_Slot_Name ("fileDialog_ListFiles");
Am_Slot_Key fileDialogPath = Am_Register_Slot_Name ("fileDialogPath");
Am_Slot_Key fileDialogDirSelected = Am_Register_Slot_Name ("fileDialogDirSelected");
Am_Slot_Key fileDialogFileSelected = Am_Register_Slot_Name ("fileDialogFileSelected");

#ifdef NEED_READDIR
/* code to simulate Berkeley directory interface for non-BSD platforms
   adapted from O'Reilly "Using C on the UNIX System" */ 

typedef struct {
  int fd;
} DIR;

static DIR current_dir;

static DIR * opendir(
   char *dir)
{
   current_dir.fd = open(dir,0);

   if (current_dir.fd >= 0) {
      struct stat stbuff;

      if (fstat(current_dir.fd, & stbuff) >= 0) {
	 if ((stbuff.st_mode & S_IFDIR) != 0) {
	   return & current_dir;
	 }
      }
      close(current_dir.fd);
   }
   return 0;
}

void closedir(DIR * dp)
{
   close(dp->fd);
}

struct direct * readdir(DIR * dp)
{
   static struct direct dir;

   do {
      if (read(dp->fd, & dir, sizeof(dir)) != sizeof(dir)) return 0;
   } while (dir.d_ino == 0);

   return(& dir);
}
#endif

static void ReadDir(char* dirName, Am_Object theDialog);

const char path_separator = '/';
const char path_separator_string[] = "/";

static void empty_list(Am_Object &the_list) 
{
  if (!(the_list.Valid)) return;

  Am_Value_List p = the_list.Get(Am_GRAPHICAL_PARTS);

  if (!(p.Valid)) return;

  for (p.End(); !p.First(); p.Prev()) {
    the_list.Remove_Part(p.Get());
  }
}

// passed to DArray:Sort
static int myCompare(char *n1, char  *n2)
{
  return strcmp(n1, n2);
}

// removes trailing path_separator, if any, from pathname
static void clean_up_pathname(char * pathname)
{
  int plen = strlen(pathname);
  if (pathname[plen - 1] == path_separator) pathname[plen - 1] = 0;
  return;
}

// removes final segment of pathname, leaving root path empty
// fails if path_separator not found
static void backup_path(char * pathname)
{
  int plen = strlen(pathname);
  int i;
  for (i = plen - 1; i >= 0; --i) {
    if (pathname[i] == path_separator) {
      // remove final segment of pathname
      pathname[i] = 0;
      return;
    }
  }
  Am_Error("file_dialog_widgets: malformed absolute pathname");
  return;
}

// appends relative pathname (filename) to absolute pathname
// fails if filename is absolute
// pathname should not end in a path_separator character
static void append_name_to_path(char * pathname, char * filename)
{
  int flen = strlen(filename);
  if (flen == 0) return;

  if (filename[0] == path_separator) {
    // filename is an absolute pathname
    Am_Error("file_dialog_widgets: need relative not absolute pathname");
    return;
  } else if (filename[0] == '.') {
    if (filename[1] == 0) {
      // current directory
      return;
    } else if (filename[1] == '.') {
      if (filename[2] == 0) {
	// parent directory
	backup_path(pathname);
	return;
      } else if (filename[2] == path_separator) {
	backup_path(pathname);
	append_name_to_path(pathname, filename + 3);
	return;
      }
    } else if (filename[1] == path_separator) {
      append_name_to_path(pathname, filename + 2);
      return;
    }
  }
  // concatenate to pathname
  strcat(pathname, path_separator_string);
  strcat(pathname, filename);
}

// returns true if append_name_to_path(pathname, filename) is a directory
// treats ".." and "." as special cases
// otherwise filename must not begin with path_separator
// or ".."#path_separator 
static bool is_dir(char * pathname, char *filename)
{
  if (filename[0] == path_separator) {
    Am_Error("file_dialog_widgets: filename must be a relative pathname");
  }
  if (filename[0] == '.') {
    if (filename[1] == 0) return true;
    if (filename[1] == '.') {
      if (filename[2] == 0) return true;
      if (filename[2] == '/') {
	Am_Error("file_dialog_widgets: filename must not start with parent directory");
      }
    }
  }

  struct stat stbuff;
  int old_len = strlen(pathname);

  append_name_to_path(pathname, filename);
  if (stat(pathname, & stbuff) >= 0) {
    if ((stbuff.st_mode & S_IFDIR) != 0) {
      pathname[old_len] = 0;
      return true;
    }
  }
  pathname[old_len] = 0;
  return false;
}

//*********************************************************************
//
//   myDirClick
//
//*********************************************************************
Am_Define_Method(Am_Object_Method, void, myDirClick, (Am_Object self))
{
  Am_Object owner = self.Get_Owner();
  Am_Object theDialog = owner.Get_Owner().Get_Owner();
  
  Am_Object selectedDir = theDialog.Peek(fileDialogDirSelected);
  
  if (selectedDir.Valid()) {
    selectedDir.Set(Am_INVERT, !selectedDir.Get(Am_INVERT));
  }

  owner.Set(Am_INVERT, !owner.Get(Am_INVERT));
  theDialog.Set(fileDialogDirSelected, owner);
}

//*********************************************************************
//
//   myDirDoubleClick
//
//*********************************************************************
Am_Define_Method(Am_Object_Method, void, myDirDoubleClick, (Am_Object self))
{
  Am_Object tmp = self.Get_Owner();
  Am_Object theDialog = tmp.Get_Owner().Get_Owner();
  Am_String s = tmp.Get(Am_TEXT);
  char * filename = s;
  char pathname[MAXPATHLEN + 1];

  strcpy(pathname, (char*) ((Am_String) theDialog.Get(fileDialogPath)));
  append_name_to_path(pathname, filename);
  theDialog.Set(fileDialogPath, pathname);

  ReadDir(pathname, theDialog);
}

//*********************************************************************
//
//   myFileClick
//
//*********************************************************************
Am_Define_Method(Am_Object_Method, void, myFileClick, (Am_Object self))
{  
  Am_Object owner = self.Get_Owner();
  Am_Object theDialog = owner.Get_Owner().Get_Owner();
  
  Am_Object selectedDir = theDialog.Peek(fileDialogFileSelected);
  
  if (selectedDir.Valid()) {
    selectedDir.Set(Am_INVERT, !selectedDir.Get(Am_INVERT));
  }

  owner.Set(Am_INVERT, !owner.Get(Am_INVERT));
  theDialog.Set(fileDialogFileSelected, owner);
}

//*********************************************************************
//
//   myFileDoubleClick
//
//*********************************************************************
Am_Define_Method(Am_Object_Method, void, myFileDoubleClick, (Am_Object self))
{  
  Am_Object tmp = self.Get_Owner();
  Am_Object theDialog = tmp.Get_Owner().Get_Owner();
  Am_Object txtSelection = theDialog.Get(fileDialog_txtSelection);
  Am_String s = tmp.Get(Am_TEXT);
  char * filename = s;
  char  pathname[MAXPATHLEN + 1];

  strcpy(pathname, (char*) ((Am_String) theDialog.Get(fileDialogPath)));
  append_name_to_path(pathname, filename);

  txtSelection.Set(Am_VALUE, pathname);
  Am_Finish_Pop_Up_Waiting(theDialog, "OK");
}

//*********************************************************************
//
//   ReadDir
//
//*********************************************************************

void ReadDir(char* dirName, Am_Object theDialog)
{
  DIR *dir = opendir(dirName);
  if (dir == NULL) {
    Am_Error("file_dialog_widgets: ReadDir cannot open directory");
  }
  
  DArray *dataArrayDirs = new DArray(MAXNAMLEN + 1);
  DArray *dataArrayFiles = new DArray(MAXNAMLEN + 1);
  struct dirent *entry;
  char pathname[MAXPATHLEN + 1];
  strcpy(pathname, dirName);
  
  while ((entry = readdir(dir))) {
    if (is_dir(pathname, entry->d_name)) {
      // directory
      dataArrayDirs->Add(entry->d_name);
    } else {
      // file (invisible files omitted)
      // to do: using the filter
      dataArrayFiles->Add(entry->d_name);
    }
  }
  closedir(dir);
  dataArrayDirs->Sort((int (*)(const void *, const void *))myCompare); 
  dataArrayFiles->Sort((int (*)(const void *, const void *))myCompare);
  
  Am_Object listDirs = theDialog.Get(fileDialog_ListDirs);
  empty_list(listDirs);
  Am_Object listFiles = theDialog.Get(fileDialog_ListFiles);
  empty_list(listFiles);
  
  Am_Object txtSelection = theDialog.Get(fileDialog_txtSelection);
  txtSelection.Set(Am_VALUE, dirName);
  
  int hLine = 0;
  int width, maxWidth = 0;
  int n = dataArrayDirs->length;
  Am_Object textLine;
  for (int i = 0; i < n; i++)  {
    textLine = Dir_Line_Proto.Create()
      .Set(Am_TEXT, dataArrayDirs->Get(i))
      .Set(Am_TOP, i * hLine)
      ;
    listDirs.Add_Part(textLine);
    
    width = textLine.Get(Am_WIDTH);
    if (width > maxWidth) maxWidth = width;
    if (hLine == 0) hLine = textLine.Get(Am_HEIGHT);  
  }
  listDirs.Set(Am_INNER_HEIGHT, n*hLine + 4);
  listDirs.Set(Am_INNER_WIDTH, maxWidth);
  
  n = dataArrayFiles->length;
  maxWidth = hLine = 0;
  int dType = theDialog.Get(fileDialogType);
  //  Am_String theFilter = txtDialog.Get_Object(fileDialog_txtFilter).Get(Am_VALUE);
  for (int i = 0; i < n; i++)  {
    if (dType == 0) textLine = Type0_File_Line_Proto.Create();
    else textLine = Am_Text.Create();

    textLine.Set(Am_TEXT, dataArrayFiles->Get(i))
      .Set(Am_TOP, i*hLine);
    listFiles.Add_Part(textLine);
    
    width = textLine.Get(Am_WIDTH);
    if (width > maxWidth) maxWidth = width;
    if (hLine == 0) hLine = textLine.Get(Am_HEIGHT);    
  }
  listFiles.Set(Am_INNER_HEIGHT, n*hLine + 4); 
  
  dataArrayDirs->Destroy();
  dataArrayFiles->Destroy();
  free(dataArrayDirs);
  free(dataArrayFiles);
}

//*********************************************************************
//
//   InitializeFileDialog
//
//*********************************************************************
Am_Object InitializeFileDialog(int dType, char *dTitle, char *dirName, char* /* fileName */)
{
  Dir_Line_Proto = Am_Text.Create("directory line prototype")
    .Add_Part(Am_One_Shot_Interactor.Create("dir click")
	      .Set(Am_START_WHEN,"LEFT_DOWN")
	      .Set(Am_DO_METHOD, myDirClick))
    .Add_Part(Am_One_Shot_Interactor.Create("dir double click")
	      .Set(Am_START_WHEN,"DOUBLE_LEFT_DOWN")
	      .Set(Am_DO_METHOD, myDirDoubleClick));
  Type0_File_Line_Proto = Am_Text.Create("type 0 file line prototype")
    .Add_Part(Am_One_Shot_Interactor.Create("file click")
	      .Set(Am_START_WHEN,"LEFT_DOWN")
	      .Set(Am_DO_METHOD, myFileClick))
    .Add_Part(Am_One_Shot_Interactor.Create("file double click")
	      .Set(Am_START_WHEN,"DOUBLE_LEFT_DOWN")
	      .Set(Am_DO_METHOD, myFileDoubleClick));

  Am_Object file_dialog = Am_Window.Create("file_dialog")
    .Set(Am_DESTROY_WINDOW_METHOD, Am_Default_Pop_Up_Window_Destroy_Method)
    .Set(Am_FILL_STYLE, Am_Motif_Gray)
    .Set(Am_TITLE, dTitle)
    //  .Set(Am_ICON_TITLE, dTitle)
    .Set(Am_WIDTH, 456)
    .Set(Am_HEIGHT, (dType > 0) ? 353 : 333)
    .Add_Part(fileDialog_txtFilter, Am_Text_Input_Widget.Create("fileDialog_txtFilter")
	      .Set(Am_LEFT, 15)
	      .Set(Am_TOP, 20)
	      .Set(Am_WIDTH, 412)
	      .Set(Am_HEIGHT, 26)
	      .Get_Object(Am_COMMAND)
	      .Set(Am_LABEL, "Filter")
	      .Get_Owner()
	      .Set(Am_FILL_STYLE, Am_Amulet_Purple)
	      )
    .Add_Part(fileDialog_txtSelection, Am_Text_Input_Widget.Create("fileDialog_txtSelection")
	      .Set(Am_LEFT, 21)
	      .Set(Am_TOP, (dType > 0) ? 260 : 240)
	      .Set(Am_WIDTH, 410)
	      .Set(Am_HEIGHT, 26)
	      .Get_Object(Am_COMMAND)
	      .Set(Am_LABEL, "Selection")
	      .Get_Owner()
	      .Set(Am_FILL_STYLE, Am_Amulet_Purple)
	      )
    .Add_Part(fileDialog_ListDirs, Am_Scrolling_Group.Create("fileDialog_ListDirs")
	      .Set(Am_LEFT, 21)
	      .Set(Am_TOP, 74)
	      .Set(Am_WIDTH, 189)
	      .Set(Am_HEIGHT, 137)
	      .Set(Am_FILL_STYLE, Am_Amulet_Purple)
	      .Set(Am_INNER_WIDTH, 400)
	      .Set(Am_INNER_HEIGHT, 400)
	      .Set(Am_H_SCROLL_BAR, 1)
	      .Set(Am_V_SCROLL_BAR, 1)
	      .Set(Am_H_SCROLL_BAR_ON_TOP, 0)
	      .Set(Am_V_SCROLL_BAR_ON_LEFT, 0)
	      )
    .Add_Part(fileDialog_ListFiles, Am_Scrolling_Group.Create("fileDialog_ListFiles")
	      .Set(Am_LEFT, 232)
	      .Set(Am_TOP, 74)
	      .Set(Am_WIDTH, 195)
	      .Set(Am_HEIGHT, 138)
	      .Set(Am_FILL_STYLE, Am_Amulet_Purple)
	      .Set(Am_INNER_WIDTH, 400)
	      .Set(Am_INNER_HEIGHT, 400)
	      .Set(Am_H_SCROLL_BAR, 1)
	      .Set(Am_V_SCROLL_BAR, 1)
	      .Set(Am_H_SCROLL_BAR_ON_TOP, 0)
	      .Set(Am_V_SCROLL_BAR_ON_LEFT, 0)
	      )
    .Add_Part(Am_Text.Create()
	      .Set(Am_LEFT, 22)
	      .Set(Am_TOP, 58)
	      .Set(Am_WIDTH, 103)
	      .Set(Am_HEIGHT, 10)
	      .Set(Am_TEXT, "Directories")
	      .Set(Am_FONT, Am_Font(Am_FONT_FIXED, true, false, false, Am_FONT_MEDIUM))
	      .Set(Am_LINE_STYLE, Am_Black)
	      .Set(Am_FILL_STYLE, Am_No_Style)
	      )
    .Add_Part(Am_Text.Create()
	      .Set(Am_LEFT, 233)
	      .Set(Am_TOP, 58)
	      .Set(Am_WIDTH, 60)
	      .Set(Am_HEIGHT, 10)
	      .Set(Am_TEXT, "Files")
	      .Set(Am_FONT, Am_Font(Am_FONT_FIXED, true, false, false, Am_FONT_MEDIUM))
	      .Set(Am_LINE_STYLE, Am_Black)
	      .Set(Am_FILL_STYLE, Am_No_Style)
    )
    .Add_Part(Am_Button_Panel.Create()
	      .Set(Am_LEFT, 22)
	      .Set(Am_TOP, (dType > 0) ? 305 : 285)
	      .Set(Am_HEIGHT, 38)
	      .Set(Am_FONT, Am_Font(Am_FONT_FIXED, true, false, false, Am_FONT_MEDIUM))
	      .Set(Am_FILL_STYLE, Am_Motif_Gray)
	      .Set(Am_LAYOUT, Am_Horizontal_Layout)
	      .Set(Am_H_SPACING, 300)
	      .Set(Am_V_SPACING, 0)
	      .Set(Am_MAX_RANK, 0)
	      .Set(Am_ITEMS, Am_Value_List()
		   .Add(Am_Standard_OK_Command.Create())
		   .Add(Am_Standard_Cancel_Command.Create())
		   )
	      )
    .Add_Part(Am_Button.Create()
	      .Set(Am_LEFT, 187)
	      .Set(Am_TOP, (dType > 0) ? 303 : 283)
	      .Set(Am_WIDTH, 76)
	      .Set(Am_HEIGHT, 38)
	      .Set(Am_FONT, Am_Font(Am_FONT_FIXED, true, false, false, Am_FONT_MEDIUM))
	      .Set(Am_FILL_STYLE, Am_Motif_Gray)
	      .Get_Object(Am_COMMAND)
	      .Set(Am_LABEL, "Filter")
	      .Get_Owner()
	      )
    .Add_Part(Am_Tab_To_Next_Widget_Interactor.Create())
    ;
  
  if (dType > 0) {
    file_dialog.Add_Part(Am_Option_Button.Create()
			 .Set(Am_LEFT, 214)
			 .Set(Am_TOP, 220)
			 .Set(Am_FILL_STYLE, Am_Motif_Gray)
			 .Set(Am_ITEMS, Am_Value_List()
			      .Add(Am_Command.Create()
				   .Set(Am_LABEL, "Option 1")
				   .Set(Am_ID, 1))
			      .Add(Am_Command.Create()
				   .Set(Am_LABEL, "Option 2")
				   .Set(Am_ID, 2))
			      )
			 )
      .Add_Part(Am_Text.Create()
		.Set(Am_LEFT, 25)
		.Set(Am_TOP, 231)
		.Set(Am_WIDTH, 190)
		.Set(Am_HEIGHT, 15)
		.Set(Am_TEXT, "Format for saved document:")
		.Set(Am_FONT, Am_Font(Am_FONT_FIXED, true, false, false, Am_FONT_MEDIUM))
		.Set(Am_LINE_STYLE, Am_Black)
		.Set(Am_FILL_STYLE, Am_No_Style)
		)
      ;
  }
  
  int x = ((int)Am_Screen.Get(Am_WIDTH) - (int)file_dialog.Get(Am_WIDTH)) / 2;
  int y = ((int)Am_Screen.Get(Am_HEIGHT) - (int)file_dialog.Get(Am_HEIGHT)) / 2;
  file_dialog.Set(Am_LEFT, x)
    .Set(Am_TOP, y);
  
  file_dialog.Add(fileDialogType, dType);
  file_dialog.Add(fileDialogPath, dirName);
  file_dialog.Add(fileDialogDirSelected, Am_NONE);
  file_dialog.Add(fileDialogFileSelected, Am_NONE);
  file_dialog.Add(Am_VALUE, Am_NONE);

  ReadDir(dirName, file_dialog);  
  
  return file_dialog;
}

//*********************************************************************
//
//   InitializeGetFileDialog
//
//*********************************************************************
Am_Object InitializeGetFileDialog(char *dirName, char *fileName)
{
  Am_Object theDialog = InitializeFileDialog(0, "Open file...", dirName, fileName );

  return theDialog;
}

//*********************************************************************
//
//   InitializePutFileDialog
//
//*********************************************************************
Am_Object InitializePutFileDialog(char *dirName, char *fileName)
{
  Am_Object theDialog = InitializeFileDialog(1, "Save file...", dirName, fileName);

  return theDialog;
}

///////////////////////////////////////////////////////////////////////////////////
//
//  Am_Get_Full_Filename
//
//   As defined in gdefs.h:
//
//   Returns the full filename of the specified name.  Depending on the
//   file system, this might add the current directory to the front.
//
//   Assumes the partial_filename is in (a subdirectory of) the current directory
//
///////////////////////////////////////////////////////////////////////////////////

extern "C" char * getwd(char *);

Am_String Am_Get_Full_Filename(Am_String partial_filename)
{
  char * filename = partial_filename;
  if (filename[0] == path_separator) {
    // already an absolute filename
    return partial_filename;
  }

  char pathname[MAXPATHLEN + 1];

  // for HP-UX, may need to use getcwd instead of getwd
  if (getwd(pathname) == 0) {
    Am_Error("Am_Get_Full_Filename unable to get current directory");
    return Am_No_String;
  }
  append_name_to_path(pathname, filename);
  return pathname;
}

///////////////////////////////////////////////////////////////////////////////////
//
//  Am_Get_Directory_Part
//
//   As defined in gdefs.h:
//
//    Pulls all the directory part out of the filename and returns it,
//    and then sets file_only_name to the file.  Either one might be
//    empty, but won't be NULL.
//
///////////////////////////////////////////////////////////////////////////////////
Am_String Am_Get_Directory_Part(Am_String filename, Am_String &file_only_name)
{
  char * name = filename;
  int len = strlen(name);
  int i;

  for (i = len; i >= 0; i--) {
    if (name[i] == path_separator) {
      name[i] = 0;
      file_only_name = name + i + 1;
      return name;
    }
  }
  file_only_name = filename;
  return "";
}

///////////////////////////////////////////////////////////////////////////////////
//
//  Am_Concat_Filename
//
//   As defined in gdefs.h:
//
//   attaches directory_part before filename (normally, just a string
//   concatenation, possibly with a separator like "/")
//
///////////////////////////////////////////////////////////////////////////////////
Am_String Am_Concat_Filename (Am_String directory_part, Am_String filename)
{
  char * fname = filename;
  char pathname[MAXPATHLEN + 1];

  strcpy(pathname, (char *) directory_part);

  clean_up_pathname(pathname);	// removes trailing path_separator if present
  append_name_to_path(pathname, fname);
  return pathname;
}

///////////////////////////////////////////////////////////////////////////////////
//
//  Am_Parse_Filename
//
//   As defined in gdefs.h:
//
//   Given a filename that might have directories, takes it apart and
//   returns the list of directories. The highest level directory (e.g.:
//   "C:" on a PC, "/" on Unix, etc.) is first in the list.  If filename
//   does not have any directory separators (either / or \ depending on
//   the system) then a value_list with one entry, the passed in
//   filename, will be returned.
//
/////////////////////////////////////////////////////////////////////////////////// 
Am_Value_List Am_Parse_Filename(Am_String filename)
{
  char * name = filename;
  char part[MAXNAMLEN];
  int len = strlen(filename);
  int name_i = 0, part_i = 0;

  Am_Value_List name_list; 

  if (name[0] == path_separator) {
    // the path_separator only appears as the root directory
    part[0] = path_separator;
    part_i = 1;
  }
  for (name_i = 0; name_i < len; name_i++) {
    if (name[name_i] == path_separator) {
      part[part_i] = 0;
      part_i = 0;
      name_list.Add(Am_String(part));
    } else {
      part[part_i] = name[name_i];
      part_i++;
    }
  }
  if (part_i != 0) {
    // last segment of filename
    part[part_i] = 0;
    name_list.Add(Am_String(part));
  }
  return name_list;
}

///////////////////////////////////////////////////////////////////////////////////
//
//  Am_Get_Directory_Contents
//
//   As defined in gdefs.h:
//
//  Returns a list of am_strings which are the names of all the files in
//  the directory.  The file named "." is not returned, but ".." is.
//
///////////////////////////////////////////////////////////////////////////////////  
Am_Value_List Am_Get_Directory_Contents(Am_String directory)
{
  char * fullname = Am_Get_Full_Filename(directory);

  DIR *dir = opendir(fullname);
  if (dir == NULL) {
    Am_Error("Am_Get_Directory_Contents cannot open directory");
  }
  
  DArray *dataArray = new DArray(MAXNAMLEN + 1);
  struct dirent *entry;
  char pathname[MAXPATHLEN + 1];
  strcpy(pathname, fullname);
  
  // read into dynamic array and sort
  while ((entry = readdir(dir))) {
    if (strcmp(entry->d_name,".")) dataArray->Add(entry->d_name);
  }
  closedir(dir);
  dataArray->Sort((int (*)(const void *, const void *))myCompare);

  int n = dataArray->length;
  Am_Value_List name_list;
  for (int i = 0; i < n; i++)  {
    name_list.Add(Am_String(dataArray->Get(i)));
  }
  dataArray->Destroy();
  free(dataArray);

  return name_list;
}

///////////////////////////////////////////////////////////////////////////////////
//
//  Am_File_Is_Directory
//
//   As defined in gdefs.h:
//
//   Returns true if the filename exists and is a directory
//
/////////////////////////////////////////////////////////////////////////////////// 
bool Am_File_Is_Directory(Am_String filename)
{
  // need absolute pathname for stat
  Am_String fullname = Am_Get_Full_Filename(filename);

  struct stat buf;
  if (stat((char*) fullname, &buf)) return false;
  if ((buf.st_mode & S_IFMT) == S_IFDIR) return true;
  return false;
}

///////////////////////////////////////////////////////////////////////////////////
//
//  Am_File_Exists
//
//   As defined in gdefs.h:
//
//   Returns true if the file exists and is not a directory
//
/////////////////////////////////////////////////////////////////////////////////// 
bool Am_File_Exists(Am_String filename)
{
  // need absolute pathname for stat
  Am_String fullname = Am_Get_Full_Filename(filename);

  struct stat buf;
  if (stat((char*) fullname, &buf)) return false;
  if ((buf.st_mode & S_IFMT) == S_IFDIR) return false;
  return true;
}

///////////////////////////////////////////////////////////////////////////////////
//
//  Am_Get_File_Size
//
//   As defined in gdefs.h:
//
//   Returns number of bytes in the file.
//
/////////////////////////////////////////////////////////////////////////////////// 
long Am_Get_File_Size(Am_String filename)
{
  // need absolute pathname for stat
  Am_String fullname = Am_Get_Full_Filename(filename);

  struct stat buf;
  if (stat((char*) fullname, &buf)) {
    Am_Error("Am_Get_File_Size: file does not exist");
  }
  else return buf.st_size;
}

///////////////////////////////////////////////////////////////////////////////////
//
//  Am_Get_File_Type
//
//   As defined in gdefs.h:
//
//   Under Unix and PC, returns the extension of the file.  If none,
//   returns NULL.  Under Mac, returns the file type as a string.  
//
///////////////////////////////////////////////////////////////////////////////////
Am_String Am_Get_File_Type(Am_String filename)
{
  char *s = filename;
  int  i, len = strlen(s);

  for (i = len - 1; i >= 0; i --) {
    if (s[i] == '.')
      return (char *) (s + i + 1);
  }
  return Am_No_String;
}

#endif
